# Standard Library
import os
import errno
import shutil

# Third Party
from azureml.core.compute import ComputeTarget, AmlCompute
from azureml.core.compute_target import ComputeTargetException
import pandas as pd
import azureml.core
from azureml.core import Workspace

from azureml.core.experiment import Experiment
from azureml.train.automl.run import AutoMLRun


def noop(*args, **kwargs) -> None:
    """
    No operation. This is a function to be used when no additional
    operation is required. This is handy in scenarios where a 
    decorator must be called, but no additional functionality is
    necessary. 

    Parameters
    ----------
    Any, all params are ignored

    Returns
    ---------
    None
    """
    pass


global logger
logger = noop


def log(message: str):
    """Takes in log strings and displays them to the configured console

    Used to pass logs to Designer

    Parameters
    ----------
    message : str
        The Message to be displayed

    """
    global logger
    logger(message)


def copy_and_create_dir(src, dest):
    try:
        shutil.copyfile(src, dest)
    except IOError as e:
        if e.errno != errno.ENOENT:
            raise
        os.makedirs(os.path.dirname(dest), exist_ok=True)
        shutil.copyfile(src, dest)


def get_workspace(
    subscription_id, resource_group, workspace_name, location="westus2"
) -> Workspace:
    """Get's an Azure ML Workspace

    Parameters
    ----------
    subscription_id : str
        The Workspace's Subscription ID
    resource_group : str
        The Workspace's Resource Group
    workspace_name : str
        The Workspace's Name
    location : str, optional
        Server location (the default is "westus2", which is in Northern California)

    Returns
    -------
    Workspace
        A Workspace object to be passed into the remaining methods
    """

    log("Attempting to find the workspace")

    try:
        ws = Workspace(
            subscription_id=subscription_id,
            resource_group=resource_group,
            workspace_name=workspace_name,
        )
        log("Workspace found: " + workspace_name)
    except:
        log("Workspace not found. Creating a new one")
        ws = Workspace.create(
            name=workspace_name,
            subscription_id=subscription_id,
            resource_group=resource_group,
            location=location,
            create_resource_group=True,
            exist_ok=True,
        )
        log("Workspace Created: " + workspace_name)
    return ws


def get_default_workspace() -> Workspace:
    """ Grabs a reference to Ashton's workspace for testing purposes"""
    return get_workspace(
        "9381ec32-2876-4c23-9bce-fa15d1478739", "Auto_ML_Testing", "Auto_ML_Test"
    )


def create_compute_resources(
    ws: Workspace, name: str, gpu: bool, min_nodes=0, max_nodes=4
) -> ComputeTarget:
    """Creates a compute resource. Deletes interfering resources.

    Creates a compute resource based on user input. Can change between GPU and CPU modes. 

    Parameters
    ----------
    ws : Workspace
        The Azure Workspace object
    name : str
        Name of the Azure Workspace
    gpu : bool
        Enable this for a GPU accelerated cluster
    min_nodes : int, optional
        Minimum number of dedicated nodes
    max_nodes : int, optional
        Maximum number of possible nodes

    Returns
    -------
    ComputeTarget
        The compute object which can be passed into a RunConfiguration object
    """

    # Use the cluster if it already exists
    try:
        cluster = ComputeTarget(workspace=ws, name=name)
        log(f'Cluster with name "{name}" differs from compute configuration.')
        log(f'Using "{name}"')
        return cluster
    except:
        pass

    if gpu:
        compute_config = AmlCompute.provisioning_configuration(
            vm_size="STANDARD_NC6",
            min_nodes=min_nodes,
            max_nodes=max_nodes,
            tags={"Type": "GPU", "min_nodes": min_nodes, "max_nodes": max_nodes},
        )
    else:
        compute_config = AmlCompute.provisioning_configuration(
            vm_size="STANDARD_D3_V2",
            min_nodes=min_nodes,
            max_nodes=max_nodes,
            tags={"Type": "CPU", "min_nodes": min_nodes, "max_nodes": max_nodes},
        )
    log("Creating New Cluster: " + name)
    cluster = ComputeTarget.create(ws, name, compute_config)
    # Removed so user doesn't have to wait for the cluster to spin up
    cluster.wait_for_completion(show_output=True)
    log("Cluster Created")

    return cluster


def get_compute_resources(
    ws: Workspace, name: str, gpu: bool, min_nodes=0, max_nodes=4
) -> ComputeTarget:
    """Get's a compute resource based on configuration.

    Try's to return a compute resource from the given values. If no resource matches, one will be created. 

    Parameters
    ----------
    ws : Workspace
        The Azure Workspace object
    name : str
        The Cluster Name
    gpu : bool
        Enable this for a GPU accelerated cluster
    min_nodes : int, optional
        The minimum number of dedicated nodes
    max_nodes : int, optional
        The maximum number of possible nodes

    Returns
    -------
    ComputeTarget
        The compute object which can be passed into a RunConfiguration object
    """

    def compare_compute_resources(
        cluster: ComputeTarget, gpu: bool, min_nodes, max_nodes
    ) -> bool:
        if gpu and cluster.tags["Type"] != "GPU":
            return False
        if not gpu and cluster.tags["Type"] != "CPU":
            return False
        if min_nodes != int(cluster.tags["min_nodes"]):
            return False
        if max_nodes != int(cluster.tags["max_nodes"]):
            return False
        return True

    try:
        cluster = ComputeTarget(workspace=ws, name=name)
        if compare_compute_resources(cluster, gpu, min_nodes, max_nodes):
            log("Exact Cluster Found: " + name)
            return cluster
    except:
        pass
    log("Exact Cluster Not Found.")
    return create_compute_resources(ws, name, gpu, min_nodes, max_nodes)


def get_generic_compute_resource(ws: Workspace, name: str) -> ComputeTarget:
    """Returns a compute resource based purely on name. Will create one if none exists.

    Parameters
    ----------
    ws : Workspace
        The Workspace Name
    name : str
        The Cluster Name

    Returns
    -------
    ComputeTarget
        The compute object which can be passed into a RunConfiguration object
    """

    try:
        cluster = ComputeTarget(workspace=ws, name=name)
        log("Generic Cluster Found: " + name)
    except:
        log("Generic Cluster Not Found. Creating a new one")
        cluster = create_compute_resources(ws, name, False)
    return cluster


def setup_project_folder(project_path: str, x: pd.DataFrame, y: pd.DataFrame):
    """Copies all necessary dependencies into a temp folder for remote runs.

    Copies the 'get_daya.py', x_train data, and y_train data into a temp folder for upload.

    Parameters
    ----------
    x : pd.DataFrame
        Training X values
    y : pd.DataFrame
        Training Y values

    Returns
    -------
    str
        Path to the temp folder
    """
    log("Copying get_data.py")
    my_path = os.path.realpath(__file__)
    src = my_path[: my_path.rfind("\\")] + "\\get_data.py"

    copy_and_create_dir(src, project_path + "\\get_data.py")
    log("Copying X and Y values")
    x.to_json(project_path + "\\x.json")
    y.to_json(project_path + "\\y.json")

    log("Project folder ready")

    return project_path


def get_most_recent_run(experiment: Experiment) -> AutoMLRun:
    """Returns the most recent AutoMLRun from an experiment.

    Parameters
    ----------
    experiment : Experiment
        The experiment where the run lives

    Returns
    -------
    AutoMLRun
        The most recent AutoMLRun
    """

    runs = experiment.get_runs(type="automl")

    most_recent = next(runs)

    while True:
        if most_recent.get_status() == "Completed":
            break
        most_recent = next(runs)

    return AutoMLRun(experiment, most_recent.id)


def convert_to_tags(dataframe: pd.DataFrame, target: str) -> dict:
    """Converts a dataframe into a dict that will be submit along with a run. Useful from schema matching
    
    Parameters
    ----------
    dataframe : pd.DataFrame
        Data from AYX
    
    Returns
    -------
    dict
        Metadata
    """
    columns = list(dataframe.columns)
    if target not in columns:
        raise Exception("Target does not exist in data")
    columns.remove(target)

    return {"target": target, "columns": columns}


def convert_from_tags(dataframe: pd.DataFrame, tags: dict) -> pd.DataFrame:
    """Converts a dataframe into a reindexed dataframe based on the tags from the train job
    
    Parameters
    ----------
    dataframe : pd.DataFrame
        Data from Alteryx
    tags : dict
        Tags from Azure
    
    Returns
    -------
    pd.DataFrame
        Reindexed Dataframe
    
    Raises
    ------
    Exception
        Invalid experiment metadata
    Exception
        Schemas do not match
    """
    import ast

    try:
        columns = ast.literal_eval(tags["columns"])
        target = tags["target"]
    except KeyError:
        raise Exception("Invalid experiment metadata")

    if target in columns:
        dataframe = dataframe.drop(columns=[target])

    try:
        dataframe = pd.DataFrame(dataframe, columns=columns)
    except:
        raise Exception("Schemas do not match")

    return dataframe
