# Standard Library
import os
import logging
import time

# Third Party
import numpy as np
import pandas as pd

from azureml.core.workspace import Workspace

# Local dependencies
from . import Azure_Utils as au


class Score:
    def __init__(self, ws: Workspace, experiment_name: str, logger: callable):
        """Instantiate an object that handles scoring

        Parameters
        ----------
        ws : Workspace
            The current workspace
        experiment_name : str
            The experiment to grab the model from
        logger : callable
            This method will be passed all logs from this class

        """

        self.ws = ws
        self.logger = logger
        self.log("Fetching Experiment")
        self.experiment = ws.experiments[experiment_name]
        self.log("Fetching Most Recent Run")
        self.ml_run = au.get_most_recent_run(self.experiment)
        self.log("Fetching the best model")
        self.best_run, self.model = self.ml_run.get_output()
        self.meta_tags = self.ml_run.get_tags()

    def log(self, message: str):
        """Pushes info logs to Designer or to the Console depending on configuration

        This is an internal helper method that provides visibility into the control flow

        Parameters
        ----------
        message : str
            Message to be displayed

        """
        self.logger(message)

    def predict(self, data: pd.DataFrame) -> pd.DataFrame:
        """Handles scoring.

        Pass this method a dataframe, and it will try to score it with this instance's model

        Parameters
        ----------
        data : pd.DataFrame
            The data to be scored

        Returns
        -------
        pd.DataFrame
            The scored data
        """

        self.log("Predicting Values")
        processed_data = au.convert_from_tags(data, self.meta_tags)
        predicted = self.model.predict(processed_data)

        output = []
        for i, value in enumerate(predicted):
            output.append(np.append(data.values[i], value))

        self.log("Converting back to a dataframe")
        return pd.DataFrame(
            output, columns=data.columns.append(pd.Index(["Prediction"]))
        )
