# Standard Library
import logging
import time
from types import SimpleNamespace
import abc

# Third Party

import azureml.core
import pandas as pd
from azureml.core import Workspace
from azureml.train.automl.run import AutoMLRun
from azureml.core.experiment import Experiment
from azureml.core.runconfig import RunConfiguration
from azureml.core.conda_dependencies import CondaDependencies
from azureml.train.automl import AutoMLConfig

# Library Specific

from . import Azure_Utils as AU


class Run(abc.ABC):
    """Run interface for ease of use with AzureML

    Don't instantiate this directly!
    """

    def __init__(
        self,
        ws: Workspace,
        model_name: str,
        task: str,
        log_func: callable,
            time_column_name="",
            max_horizon=50,
        iterations=10,
        metric="accuracy",
        max_cores=1,
        max_iterations=1,
        cross_validations=3,
        iteration_timeout=10,
        experiment_timeout=60,
    ):
        """Setup a Run


        Parameters
        ----------
        ws : Workspace
            The Workspace where the model will run
        model_name : str
            Desired model name
        task : str
            Either 'classification' or 'regression'
        log_func: callable
            A function that will be called whenever a log message needs to be displayed
        iterations : int, optional
            Number of iterations to train (the default is 10, which is a good test size)
        metric : str, optional
            Metric to train by (the default is "accuracy", which is generally a good one. You can also use 'AUC_Weighted')
        max_cores : int, optional
            Max number of cores to use per iteration (the default is 1)
        max_iterations : int, optional
            Max concurrent iterations (the default is 1)
        cross_validations : int, optional
            Number of cross validations (the default is 3)
        iteration_timeout: int, optional
            Max minutes an iteration can run for (the default is 10)
        experiment_timeout: int, optional
            Max minutes an experiment can run for (the default is 60)

        """
        self.logger = log_func
        AU.logger = log_func

        self.log("Creating Experiment")

        self.experiment = Experiment(ws, model_name)
        self.task = task

        self.log("Experiment Created: " + self.experiment.get_portal_url())

        self.log("Configuring AutoML Settings")

        self.automl_settings = {
            "iteration_timeout_minutes": iteration_timeout,
            "experiment_timeout_minutes": experiment_timeout,
            "iterations": iterations,
            "preprocess": True,
            "verbosity": logging.INFO,
            "n_cross_validations": cross_validations,
            "max_cores_per_iteration": max_cores,
            "max_concurrent_iterations": max_iterations,
        }

        if task == "forecasting":
            self.automl_settings["time_column_name"] = time_column_name
            self.automl_settings["max_horizon"] = max_horizon
            self.automl_settings["preprocess"] = False

        self.automl_config = None
        self.data = SimpleNamespace(x=None, y=None)

        self.meta_tags = None

    def log(self, message: str):
        """Pushes info logs to Designer or to the Console depending on configuration
        
        This is an internal helper method that provides visibility into the control flow
        
        Parameters
        ----------
        message : str
            Message to be displayed
        
        """

        self.logger(message)

    def set_data(self, dataframe: pd.DataFrame, target_var: str):
        """Prepares a pandas dataframe for testing.

        Parameters
        ----------
        dataframe : pd.DataFrame
            The training data
        target_var : str
            The Column name that you want to predict
        test_size : float, optional
            What percentage(0.0-1.0) of the data should be reserved for testing (the default is 0.2)
        random_state : int, optional
            This is the seed for shuffling the daa (the default is 123 just for fun)

        """
        self.log("Recieved Data")
        self.meta_tags = AU.convert_to_tags(dataframe, target_var)
        x_train = dataframe.drop(target_var, axis=1)
        y_train = dataframe[[target_var]]
        self.log("Split X and Y")
        self.data.x = x_train
        self.data.y = y_train

    @abc.abstractmethod
    def run(self, path: str) -> AutoMLRun:
        """This method will run the training and return back an AutoMLRun object.

        Parameters
        ----------
        path : str
            The temp path to run the experiment from.

        Returns
        -------
        AutoMLRun
            This object contains references to all of the generated models
        """

        pass


class LocalRun(Run):
    def run(self) -> AutoMLRun:
        """This method will run the training locally and return back an AutoMLRun object.

        Returns
        -------
        AutoMLRun
            This object contains references to all of the generated models
        """
        self.log("Creating Project Folder")
        project_folder = AU.setup_project_folder(self.data.x, self.data.y)

        self.automl_config = AutoMLConfig(
            task=self.task,
            path=project_folder,
            debug_log="automl_errors.log",
            X=self.data.x.values,
            y=self.data.y.values.flatten(),
            **self.automl_settings
        )

        self.log("Starting Local Run")

        run = self.experiment.submit(self.automl_config, show_output=False)
        return run


class RemoteRun(Run):
    def set_compute(
        self,
        ws: Workspace,
        compute_name: str,
        generic: bool,
        gpu: bool,
        min_nodes=0,
        max_nodes=4,
    ):
        """Set's the compute object for a remote run.

        If no compute object is found, one will be created.

        Parameters
        ----------
        ws : Workspace
            The Workspace for the experiment
        compute_name : str
            Your cluster name
        generic : bool
            Enable this if the next three parameters don't matter
        gpu : bool
            Enable this for GPU acceleration
        min_nodes : int, optional
            The minimum number of dedicated nodes (the default is 0)
        max_nodes : int, optional
            The maximum number of possible nodes (the default is 4)

        """

        if generic:
            self.log("Finding Generic Compute Resource")
            compute = AU.get_generic_compute_resource(ws, compute_name)
        else:
            self.log("Finding Specific Compute Resource: " + compute_name)
            compute = AU.get_compute_resources(
                ws, compute_name, gpu, min_nodes=min_nodes, max_nodes=max_nodes
            )

        if compute.tags["Type"] == "GPU":
            self.log("Enableing GPU Acceleration")
            base_image = azureml.core.runconfig.DEFAULT_GPU_IMAGE
        else:
            base_image = azureml.core.runconfig.DEFAULT_CPU_IMAGE

        self.log("Creating Docker Image")

        self.conda_run_config = RunConfiguration(framework="python")

        self.conda_run_config.target = compute
        self.conda_run_config.environment.docker.enabled = True
        self.conda_run_config.environment.docker.base_image = base_image

        self.log("Creating Dependencies")

        cd = CondaDependencies.create(
            pip_packages=["pandas", "numpy"]
        )
        self.conda_run_config.environment.python.conda_dependencies = cd

    def run(self, path: str) -> AutoMLRun:
        """This method will run the training remotely and return back an AutoMLRun object.

        Parameters
        ----------
        path : str
            The temp path to run the experiment from.

        Returns
        -------
        AutoMLRun
            This object contains references to all of the generated models
        """

        self.log("Creating Project Folder")
        AU.setup_project_folder(path, self.data.x, self.data.y)

        self.automl_config = AutoMLConfig(
            task=self.task,
            debug_log="automl_errors.log",
            path=path,
            run_configuration=self.conda_run_config,
            data_script=path + "\\get_data.py",
            **self.automl_settings
        )

        self.log("Submitting Remote Run")

        run = self.experiment.submit(
            self.automl_config, show_output=False, tags=self.meta_tags
        )

        import webbrowser
        webbrowser.open_new_tab(self.experiment.get_portal_url())

        while True:
            status = run.get_status()
            if status in [
                "NotResponding",
                "Canceled",
                "Failed",
                "Completed",
                "CancelRequested",
            ]:
                break

            metrics = run.get_metrics()
            if len(metrics.keys()) != 0:
                self.log(str(metrics))
            else:
                self.log(status)

            time.sleep(30)

        # shutil.rmtree(project_folder) # Engine should handle this
        return run
