# Built in Libraries
import sys
import logging
from os.path import dirname, abspath, join
from importlib import import_module
import logging

# 3rd Party Libraries
import pickle
import pandas as pd

# Alteryx Libraries
import AlteryxPythonSDK as Sdk

# Custom libraries
from snakeplane.plugin_factory import PluginFactory

# Initialization of the plug in factory, used for making the AyxPlugin class
factory = PluginFactory("AzureML_Train")


@factory.initialize_plugin
def init(workflow_config, user_data, logger):
    # Grab UI Data
    user_data.target = workflow_config.get("target")
    user_data.task = workflow_config.get("task")
    user_data.model_name = workflow_config.get("name")
    user_data.iterations = int(workflow_config.get("iterations"))
    user_data.concurrent = int(workflow_config.get("concurrent"))
    user_data.compute_name = workflow_config.get("computeName")
    user_data.min_nodes = int(workflow_config.get("minNodes"))
    user_data.max_nodes = int(workflow_config.get("maxNodes"))
    user_data.gpu = workflow_config.get("gpu") == "True"
    user_data.max_cores = int(workflow_config.get("maxCores"))
    user_data.max_iteration_time = int(workflow_config.get("maxTimePerIteration"))
    user_data.max_run_time = int(workflow_config.get("maxRunTime"))

    user_data.cross_validations = int(workflow_config.get("crossValidations"))

    user_data.subscription_id = workflow_config.get("subscription_id")
    user_data.resource_group = workflow_config.get("resource_group")
    user_data.workspace_name = workflow_config.get("workspace_name")

    user_data.time_column = workflow_config.get("timeColumn")
    user_data.max_horizon = int(workflow_config.get("maxHorizon"))

    initSuccess = True

    # Check for required Values
    if user_data.model_name is None:
        logger.display_error_msg("Model requires a name")
        initSuccess = False

    if user_data.compute_name is None:
        logger.display_error_msg("Must specify a compute resource")
        initSuccess = False

    if user_data.target is None:
        logger.display_error_msg("Select a target field")
        initSuccess = False

    if user_data.task is None:
        logger.display_error_msg("Select a model type")

    if user_data.subscription_id is None:
        logger.display_error_msg("Subscription ID is Required")
        initSuccess = False

    if user_data.resource_group is None:
        logger.display_error_msg("Resource Group is Required")
        initSuccess = False

    if user_data.workspace_name is None:
        logger.display_error_msg("Workspace Name is Required")
        initSuccess = False

    if user_data.task == "forecasting":
        if user_data.time_column is None:
            logger.display_error_msg("You must select a time column when using forecasting")
            initSuccess = False

    return initSuccess


@factory.process_data(mode="batch", input_type="dataframe")
def process_data(input_mgr, output_mgr, user_data, logger):
    logging.setLoggerClass(logging.Logger)
    try:
        # Import on processing
        from AzureML.train import RemoteRun, LocalRun
        from AzureML import Azure_Utils as AU

        # Pass in the data
        data_in = input_mgr["Data"][0]
        data = data_in.data

        # Instantiate Workspace
        ws = AU.get_workspace(
            user_data.subscription_id,
            user_data.resource_group,
            user_data.workspace_name,
        )

        # Instantiate Run
        run = RemoteRun(
            ws,
            user_data.model_name,
            user_data.task,
            logger.display_info_msg,
            iterations=user_data.iterations,
            max_iterations=user_data.concurrent,
            max_cores=user_data.max_cores,
            cross_validations=user_data.cross_validations,
            iteration_timeout=user_data.max_iteration_time,
            experiment_timeout=user_data.max_run_time,
            time_column_name=user_data.time_column,
            max_horizon=user_data.max_horizon
        )

        # Pass in data to the run
        run.set_data(data, user_data.target)

        # Setup Compute Resources
        run.set_compute(
            ws,
            user_data.compute_name,
            False,
            user_data.gpu,
            min_nodes=user_data.min_nodes,
            max_nodes=user_data.max_nodes,
        )

        # Create a temp folder
        file_path = output_mgr._plugin._engine_vars.alteryx_engine.create_temp_file_name(
            extension="txt", options=0
        )
        project_path = file_path[: file_path.rfind("\\")] + "\\train"

        # Submit the run
        output = run.run(project_path)

    except Exception as e:
        logger.display_error_msg(str(e))


AyxPlugin = factory.generate_plugin()
